;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;
; PureBasic Network Client Version .01
;
; Author : Karbon (AKA Mitch) - Many thanks to Tranquil as lots of this was based off examples he sent me.
;
; Date   : July 1st, 2003
;
; License : None, do with this as you please.
;
; Warranty : Absolutely none. This code indeed blow your computer up and burn your house down. I'm not 
;            to blame if it does that or anything else you don't want it to!
;
; Contact : mitch@purebasic.org or 37955574 on ICQ
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

IncludeFile "client_Windows.pb"

;
; For capturing our network events inside our callback
;

#NETWORK_EVENT = #WM_User + 1

#BUFFER_SIZE = 60000

;
; Global socket descriptor
;

Global socket.l

; Pointer to our receive buffer where we store incoming data until it's parsed.
;

*p_receive_buffer = AllocateMemory(0, #BUFFER_SIZE)

;
; my_send(text_to_send_to_server$)
;
; Arguments : 1 string argument containing text to send out (which limits it to 64k as of PB 3.7)
;
; Returns : Number of bytes sent (as a long value)
;           
; Description :
;
; Wraps up, checksums and pushes out all of send_text$ it can through the server 
;

Procedure my_send(send_text$)

  ;
  ; Remember to include space for the end of line character
  ;
  text_length = Len(send_text$)+1

  ;
  ; Pointer to memory buffer for our send text
  ;

  *p_send_buffer.l = AllocateMemory(1, text_length)

  ;
  ; Add 4 bytes for our integer indicating length of data
  ; and 4 bytes for our CRC32 checksum data
  ;
  text_length + 8

  ;
  ; Store the size of our data in the first 4 bytes of the data we send (so we 
  ; know how much to expect on the other end).
  ;
  PokeL(*p_send_buffer,text_length)
  
  ;
  ; Skip the first 4 bytes in our buffer (that's where we put the data size!)
  ; and stick our whole string in there. 
  ;
  
  PokeS(*p_send_buffer + 4,send_text$)

  ;
  ; Now we calculate a CRC checksum of the data so the server knows when something
  ; gets corrupted.
  ;
  ; We only want to checksum the data not part of the size or CRC, that's the 
  ; reason for moving the buffer around.
  ;

  crc_checksum.l = CRC32Fingerprint(*p_send_buffer + 4,text_length - 8)
  
  ; Stick the checksum in there at the end.
  
  PokeL(*p_send_buffer + text_length - 4, crc_checksum)

  ;
  ; Now we look until send_() returns the number of bytes we wanted to send.
  ; This could take several tries.
  ;

  bytes_sent = 0
  
  total_bytes_sent = 0
  
  Repeat
  
  ;
  ; send_() will return the number of bytes it actually sent.
  ;
  
  bytes_sent.l = send_(socket , *p_send_buffer , text_length ,0)
  
  total_bytes_sent + bytes_sent
  
  ;
  ; Ok, so sent returned an error. Not good.
  ;
  
  If bytes_sent = #SOCKET_ERROR
  
    wsa_error = wsagetlasterror_()
    
    Select wsa_error
    
      Case #WSANOTINITIALISED
      
        error_text$ = "WSA Not initialized"
        
      Case #WSAENETDOWN
      
        error_text$ = "Network subsystem failed!"
        
      Case #WSAENOTCONN 
      
        error_text$ = "No connection detected"
        
      Case #WSAEINTR
      
        error_text$ = "Blocking call canceled with WSACancelBlocking"
        
      Case #WSAEINPROGRESS
      
        error_text$ = "Socket already in use, unable to send"
        
      Case #WSAENOTSOCK
      
        error_text$ = "Not a socket!"
        
      Case #WSAESHUTDOWN
      
        error_text$ = "Socket closed."
        
      Case #WSAEOPNOTSUPP
      
        error_text$ = "OOB deteced but socket is not SOCK_STREAM type"
        
      Case #WSAEMSGSIZE
      
        error_text$ = "Datagram too big, it has been truncated"
        
      Case #WSAEINVAL
      
        error_text$ = "Socket not bound."
        
      Case #WSAECONNABORTED
      
        error_text$ = "Connection aborted due to timeout or other failure."
        
      Case #WSAECONNRESET
      
        error_text$ = "Server closed the connection."
        
      Case #WSAEWOULDBLOCK
      
        bytes_sent = 0
        
        total_bytes_sent = 0
        
        Delay(250)
    
    EndSelect
  
  EndIf ; = SOCKET_ERROR
  
  ;
  ; Loop until we're finished sending or we come up with an error.
  ;
  
  Until total_bytes_sent >= text_length Or bytes_sent = #SOCKET_ERROR
  
  Debug "Total sent: "+Str(total_bytes_sent)
  
  FreeMemory(1)
  
  ProcedureReturn total_bytes_sent
  
EndProcedure

;
; Next is our callback function that handles all network events.
;
; The reason for using a callback is that the PB waitwindowevent() might miss some
; events (specifically network events!) if the window is being resized, moved or in 
; other ways changed.
; 
; Using a callback insures we'll get all the events (messages) we want!
;
    
Procedure my_window_callback(WindowID, Message, wParam, lParam) 

  ;
  ; We do this so PB can handle all the other events later in the 
  ; WaitWindowEvent() loop
  ;
  
  result.l = #PB_ProcessPureBasicEvents 
  
  ;  
  ; Message is passed in by windows, we specified what event number to give network events
  ; in the WSAAsyncSelect_() call below.
  ;
    
  If Message = #NETWORK_EVENT

    ;
    ; NetworkClientEvent() is documented as returing the following values:
    ;
    ; 0 : Nothing has happened 
    ; 2 : Raw Data has been received (To be Read with ReceiveNetworkData()) 
    ; 3 : A file has been received (To be Read with ReceiveNetworkFile()) 

    client_event.l = NetworkClientEvent(socket.l)
    
    Select client_event
    
      Case 2
      
        Debug "Data received"

        If bytes_expected > 0
        
          bytes_received = recv_(socket , *p_receive_buffer + total_bytes_received , bytes_expected - total_bytes_recieved, 0)
      
          ;
          ; No real error handling for receiving (it's taken care of on the server side)
          ;
      
          If bytes_received <> #SOCKET_ERROR
          
            total_bytes_received + bytes_received
            
            ;
            ; Remember we tagged the size of the full amount of data into the first 4 
            ; bytes of what is sent.
            ;
            
            If total_bytes_received = PeekL(*p_receive_buffer)
            
              ;
              ; Transmission complete, now we deal with the data sent to us
              ;
              ; Remember to skip the first 4 bytes, it just contains the size of the data we sent.
              ;
            
              command_text$ = PeekS(*p_receive_buffer + 4)
              
              ; Reset these values and listen for more
              
              total_bytes_received = 0
              
              bytes_expected = -1
                          
              ;
              ; Here we've received everything the server says we should have, and are free to do 
              ; something with the data.
              ;
              ; Note that in this example the server isn't sending anything back to the client. 
              ;
              
              
            EndIf ; total_bytes_received
            
          EndIf ; <> #SOCKET_ERROR
          
        EndIf ; bytes_expected > 0
      
        ;
        ; This insures we get the full first 4 bytes (the size of the data coming in) before we do anything.
        ;
        
        If total_bytes_received < 3 And bytes_expected = 0
        
          bytes_received = recv_(socket,*receivebuffer+totalreceived,4-totalreceived,0)   ; Receive TCP/IP Buffer (max. 4 Bytes)
          
          If bytes_received <> #SOCKET_ERROR
          
            ;
            ; First 4 bytes contain a long indicating the size of the rest of the data
            ;
            
            If bytes_received = 4
              
              bytes_expected.l = PeekL(*p_receive_buffer)
              
            EndIf
            
            If bytes_received > 0
            
              total_bytes_received + bytes_received
            
            EndIf ; bytes_received > 0
            
          EndIf ; <> SOCKET_ERROR
          
        EndIf ; total_bytes_received < 3
              
        If bytes_expected = -1
          
          receiveexpected = 0
         
        EndIf

    EndSelect
  
  EndIf ; NETWORK_EVENT
  
  ProcedureReturn result
    
EndProcedure

;
; Open the window and point to our callback where network events are handled.
;

window_handle.l = Window_wnd_windowmain()

SetWindowCallback(@my_window_callback())

; Init the network

InitNetwork()

;
; Pop open the connection, if failed then exit.
;

socket.l = OpenNetworkConnection("127.0.0.1",6000)

If socket = 0 

  MessageRequester("Error", "Can't make network connection !", 0)

  End
  
EndIf

;
; Make sure windows reports everything we care about as a #NETWORK_EVENT
;

WSAAsyncSelect_(socket, window_handle ,#NETWORK_EVENT,#FD_READ | #FD_Write |#FD_OOB)

;
; Our main event loop
;
; Note that network events are handled in the calback above
;

Repeat
  
  EventID=WaitWindowEvent()
    
  Select EventID
    
    Case #PB_EventGadget
        
      Select EventGadgetID()
        
        Case #Gadget_wnd_windowmain_btn_send
        
;            send_server_message()      
                  
;           If EventType() = #PB_EventType_LeftClick
;                           
;             DisplayDetail()
;             
;           EndIf

            send_string$ = "command"+Chr(4)+"dootme"+Chr(4)+"customer_id"+Chr(4)+"31337"

            my_send(send_string$)
         
      EndSelect
         
    EndSelect

Until EventID = #PB_EventCloseWindow

;
; Even though this is done automatically, we do it anyway
;

FreeMemory(0)

End
; ExecutableFormat=Windows
; CursorPosition=365
; FirstLine=1
; EOF