;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;
; PureBasic Network Server Version .01
;
; Author : Karbon (AKA Mitch) - Many thanks to Tranquil as lots of this was based off examples he sent me.
;
; Date   : July 1st, 2003
;
; License : None, do with this as you please.
;
; Warranty : Absolutely none. This code indeed blow your computer up and burn your house down. I'm not 
;            to blame if it does that or anything else you don't want it to!
;
; Contact : mitch@purebasic.org or 37955574 on ICQ
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

IncludeFile "server_Windows.pb"

IncludeFile "nvpair.pb"

;
; Declarations for procedures that need to be used before they really exist
;

Declare find_free_user_slot()

Declare find_user_slot(socket_id.l)

Declare parse_commands(command_string$)

; Port for our server to listen on

#PORT_NUM = 6000

;
; Bank for our local memory buffer in the my_send() function
;

#LOCALBANK = 1000

; -----------------------------------------
; Global / Constant Variables and Structures
; -----------------------------------------

Structure User
 
  ; Socket ID of user
  
  user_socket.l
  
  ; Pointer To Receivebuffer
  
  *p_receive_buffer.l

  ; Counter for received bytes
    
  bytes_received.l

  ; This is the amount of data we'll expect (and require)
  
  expected_bytes.l
    
EndStructure

;
; Max number of users we support
; In theory this is only limited by the amount of RAM you have
; and the numbero f available file descriptors in Windows (your guess is as good as mine!)
;

#MAXUSERS = 100

;
; Number of bytes for our buffer(s)
;

#BUFFERSIZE = 12000

;
; For network events in the callback procedure
;

#NETWORKEVENT = #WM_User+1

Global server_socket.l

;
; Nice little array for our users.
;

Dim a_users.User(#MAXUSERS)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                   Start our various procedures here.                                                              ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; This is our window callback procedure. Using this insures we won't lose network events if the server window
; is resized or otherwise bothered.
;

Procedure my_window_callback(WindowID, Message, wParam, lParam) 

   result.l = #PB_ProcessPureBasicEvents 

    ;
    ; Handle Networkevents here
    ;
    ; #NETWORKEVENT is defined above and passed to a WSAAsyncSelect_() call
    ; in our start_server() procedure to indicate a network event of the type we specify
    ;
         
    If Message = #NETWORKEVENT
      
      ; NetworkServerEvent returns these:
      ;
      ;   0 : Nothing has happened 
      ;   1 : A new client has been connected To the server 
      ;   2 : Raw Data has been received (To be Read with ReceiveNetworkData()) 
      ;   3 : A file has been received (To be Read with ReceiveNetworkFile()) 
      ;   4 : A client has quit the server (deconnection) 

      network_event.l = NetworkServerEvent()

      Select network_event
        
        ;
        ; On Connection we find the user a slot in our connection array
        ;
        Case 1
        
          ;
          ; You should code error checking in here to see if slot is < 0 
          ; if it is, then you're out of room in your a_users array and 
          ; need To take some sort of action to notify the user, etc.
          ;
          
          slot.l = find_free_user_slot()

          ;          
          ; This is what's used to reference the connection EVERYWHERE, so it's important.
          ;
          
          a_users(slot)\user_socket = NetworkClientID()

          ;
          ; And the memory To use as the connection's receive buffer..
          ;
                              
          a_users(slot)\p_receive_buffer = AllocateMemory(slot,#BUFFERSIZE)
                      
          Debug "Client connected, using slot "+Str(slot)
        
        ; 
        ; User disconnected. Again you'll probably want to do more here for a *real* server.
        ;
        
        Case 4          
        
          ;  
          ; Figure out which slot this user was in..
          ;
            
          slot = find_user_slot(NetworkClientID())
          
          ;
          ; Reset the socket and free the buffer memory
          ;
          
          a_users(slot)\user_socket = 0
          
          FreeMemory(slot)
                      
        ;
        ; Raw data received
        ;
        ; Pay attention, it gets a little crazy in here.
        ;
        
        Case 2

          ;
          ; We have to find out the slot for the user who send us the datas
          ;
          
          slot.l = find_user_slot(NetworkClientID())

          Debug "New data received at server, user slot "+ Str(slot)
                    
          ;
          ; We wait until we know how much data to expect before continuing here. 
          ;
          ; We know how much to expect from the first 4 bytes coming from the client (a long variable)
          ;
          
          If a_users(slot)\expected_bytes > 0
          
            ;
            ; Point to the user's receive buffer
            ;
            
            *p_receive_buffer = a_users(slot)\p_receive_buffer
            
            ;
            ; Get as much data as we need
            ;
            
            local_received.l = recv_(NetworkClientID(), *p_receive_buffer + a_users(slot)\bytes_received , a_users(slot)\expected_bytes - a_users(slot)\bytes_received,0)
            
            ;
            ; Error checking on the socket
            ;
            
            If local_received <> #SOCKET_ERROR 

              a_users(slot)\bytes_received + local_received
             
               If a_users(slot)\bytes_received = a_users(slot)\expected_bytes
               
                  ;
                  ; So here we've received everything the client says we should have, now 
                  ; we have to do something with that data. If you're using this as an example 
                  ; then this is the point where you will start customizing heavily.
                  ;
                  ; All I do is parse the commands (in my case, name/value pairs separated by chr(4) chars)
                  ; into the linked list of structs in the NVPairs.pb file. From that I can make decisions
                  ; based on what the client sent me. You can choose to do the same, or do it your own way :-)
                  ;
                  
                  ; Grab the command string from our buffer
                  
                  command_string$ = PeekS(*p_receive_buffer + 4)

                  ;
                  ; This parses the commands into name/value pairs (accessed through av_* functions)
                  ;
                                     
                  parse_commands(command_string$)
                  
                  ; All our expected data was received, so we reset and wait for more!
                  
                  a_users(slot)\bytes_received = 0
                  
                  ; This makes sure we wait for the full header before continuing with the read
                  
                  a_users(slot)\expected_bytes = -1
                                  
              EndIf ; l_bytes_exected 
            
            Else
            
              ;
              ; Socket error recovery here. Personally I haven't decided what I'm going to do
              ; when a socket error occurs, so I'll leave that as an exercise :-)
              ;
              
              Debug "Socket Error!!!"            

            EndIf ; SOCKET_ERROR 
          
          EndIf ; expected_bytes
          
          ;
          ; This is where we determine how much data to expect.
          ;
          ; We get the first 4 bytes that indicates how much data is being sent and
          ; set the sizes depending on that value..
          ;
      
          If a_users(slot)\bytes_received < 3 And a_users(slot)\expected_bytes = 0
            
            ;
            ; Point to our user's buffer
            ;
            
            *p_receive_buffer = a_users(slot)\p_receive_buffer
            
            ;
            ; Grab 4 bytes
            ;
            
            local_received = recv_(NetworkClientID(),*p_receive_buffer + a_users(slot)\bytes_received,4 - a_users(slot)\bytes_received,0)
            
            ;
            ; Error checking for the socket and set our expected bytes value based on what we read from the socket, 
            ;
            
            If local_received <> #SOCKET_ERROR

              If local_received = 4
                
                ;
                ; Bada bing, we have our 4 byte header and can now tell how much data is coming in this round.
                ;
                
                a_users(slot)\expected_bytes = PeekL(*p_receive_buffer)
              
              EndIf
              
                ;
                ; Remember we track bytes received so we have to count these 4!
                ;
                
                a_users(slot)\bytes_received + local_received

            Else 
            
              ;
              ; Socket error recovery here. Personally I haven't decided what I'm going to do
              ; when a socket error occurs, so I'll leave that as an exercise :-)
              ;
            
              Debug "Socket error!!"

            EndIf

          EndIf

          ;
          ; If expected_bytes is -1 that means we're waiting for the full header to come in
          ;
          ; We reset this value so we can check again in our loop.
          ;
          
          If a_users(slot)\expected_bytes = -1
            
            a_users(slot)\expected_bytes = 0
           
          EndIf
          
        EndSelect
        
    EndIf ; #NETWORKEVENT check

    ;
    ; Throw this back to PB so the waitwindowevent() stuff will work..
    ;
    
    ProcedureReturn result
    
EndProcedure

;
; parse_commands(command_string_from_where_ever$)
;
; Arguments : 1 string value (which limits it to 64k as of PB 3.7)
;
; Returns : Nothing.
;
; Description :
;
; Parses out our commands from the string sent from the client.
; 
; I'm using chr(4) to separate name/value pairs, you can do it your own way if you'd like.
;

Procedure parse_commands(command_string$)

    Debug "parser caught "+ command_string$

    ; Kill our old list (if there was one) and make way for the new data!
    
    nv_clear()

    ;
    ; We have to break up the command string into name/value pairs for use.
    ; 
 
    length.l = Len(command_string$)
    
    processed.l = 0
    
    index.l = 1
    
    While processed < length
    
      name$ = StringField(command_string$, index, Chr(4))
      
      index + 1
      
      value$ = StringField(command_string$, index, Chr(4))      
      
      ; Record the number of chars we've processed so we know when to quit!
      
      processed + (Len(name$) + Len(value$) + 2)

      ; Store the pair(s) in our list.
      
      nv_add_pair(name$ , value$)
      
      index + 1
      
    Wend
    
    ;
    ; For debugging only
    ;
    
    nv_list_pairs()
    
EndProcedure

;
; find_user_slot(valid_socket_id.l)
;
; Arguments : 1 long value ( from NetworkClientID() )
;
; Returns : index of the user on socket_id in the a_users array 
;
; Description :
;
; Searches through the array and lets us know where a given user is
; so that we can reference their receive buffer, socket and other information.
;

Procedure find_user_slot(socket_id.l)

  result = -1

  count = 0
  
  Repeat

    ;
    ; We found our user so we set the result to the index of the 
    ; array that we were working on.
    ;
    
    If a_users(count)\user_socket = socket_id

      result = count

    Else

      count + 1

    EndIf

  Until count = #MAXUSERS Or result => 0

  ProcedureReturn result

EndProcedure


;
; find_free_user_slot()
;
; Arguments : Nothing
;
; Returns : On success : index indicating the first unused slot in the array
;           On error : value less than zero (indicating the user array is full)
;
; Description :
;
; Finds an unused slot in our a_users array for the new user, and returns an index 
; indicating placement in the a_users array
;

Procedure find_free_user_slot()

  result = -1

  count = 0
  
  Repeat

    ; Zero indicates a free slot
    
    If a_users(count)\user_socket = 0
 
      result = count
      
    Else

      count + 1

    EndIf

  Until count = #MAXUSERS Or result >= 0

  ProcedureReturn result

EndProcedure

;
; start_server()
;
; Arguments : Nothing
;
; Returns : Nothing
;           
; Description :
;
; Just a wrapper function to start the network server listening on #PORT_NUM
; Throws an error and ends the program if it fails.
;

Procedure start_server()

  server_socket.l = CreateNetworkServer(#PORT_NUM)

  ; Zero is failure, non-zero success
  
  If server_socket.l = 0 
  
    MessageRequester("","Could not initialize WSA.",0)
  
  Else
  
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ; VERY IMPORTANT!
    ;
    ; When doing network events in a call back it's important to let windows 
    ; know what events to call #NETWORKEVENTS.
    ;
    ; In this case we call #FD_ACCEPT or #FD_READ or #FD_CLOSE a #NETWORKEVENT
    ;
    ; Look at the callback function (my_window_callback()) and see how the #NETWORKEVENT is used
    ;
    ; NOTE that this is different in the client.
    ;
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    WSAAsyncSelect_(server_socket, WindowID(), #NETWORKEVENT, #FD_ACCEPT|#FD_READ|#FD_CLOSE) 

    Debug "Server initialized and listening on port "+Str(#PORT_NUM)
  
  EndIf
  
EndProcedure


;
; stop_server()
;
; Arguments : Nothing
;
; Returns : Nothing
;           
; Description :
;
; Just a wrapper function to stop the network server listening on #PORT_NUM
;

Procedure stop_server()

  ;
  ; Might need to do more here eventually.
  ;

  CloseNetworkServer()
  
EndProcedure

;
; my_send(send_socket.l,send_text$)
;
; Arguments : 1 long argument indicating the socket to which to send the information. This would be the 
;              NetworkClientID() of the user that you want to receive the data.
;             1 string argument containing text to send out (which limits it to 64k as of PB 3.7)
;
; Returns : Number of bytes sent (as a long value)
;           
; Description :
;
; Wraps up, checksums and pushes out all of send_text$ it can through the server 
;

Procedure my_send(send_socket.l , send_text$)

  ;
  ; Local memory buffer for our send text
  ;

  *p_send_buffer.l = AllocateMemory(#LOCALBANK, text_length)

  ;
  ; Remember to include space for the end of line character
  ;
  text_length.l = Len(send_text$)+1

  ;
  ; Add 4 bytes for our integer indicating length of data
  ; and 4 bytes for our CRC32 checksum data
  ;
  text_length + 8

  ;
  ; Store the size of our data in the first 4 bytes of the data we send (so we 
  ; know how much to expect on the other end).
  ;
  PokeL(*p_send_buffer,text_length)
  
  ;
  ; Skip the first 4 bytes in our buffer (that's where we put the data size!)
  ; and stick our whole string in there. 
  ;
  
  PokeS(*p_send_buffer + 4,send_text$)

  ;
  ; Now we calculate a CRC checksum of the data so the server knows when something
  ; gets corrupted.
  ;
  ; We only want to checksum the data not part of the size or CRC, that's the 
  ; reason for moving the length arg around.
  ;

  checksum.l = CRC32Fingerprint(*p_send_buffer + 4,text_length - 8)
  
  ; Stick the checksum in there at the end.
  
  PokeL(*p_send_buffer + text_length - 4, checksum)

  ;
  ; Now we look until send_() returns the number of bytes we wanted to send.
  ; This could take several tries.
  ;

  bytes_sent.l = 0
  
  total_bytes_sent.l = 0

  retry_send.l = 0
    
  Repeat
  
  ;
  ; send_() will return the number of bytes it actually sent.
  ;
  
  bytes_sent.l = send_(send_socket , *p_send_buffer , text_length ,0)
  
  total_bytes_sent + bytes_sent
  
  ;
  ; Ok, so sent returned an error. Not good.
  ;
  
  If bytes_sent = #SOCKET_ERROR
  
    wsa_error = wsagetlasterror_()
    
    Select wsa_error
    
      Case #WSANOTINITIALISED
      
        error_text$ = "WSA Not initialized"
        
      Case #WSAENETDOWN
      
        error_text$ = "Network subsystem failed!"
        
      Case #WSAENOTCONN 
      
        error_text$ = "No connection detected"
        
      Case #WSAEINTR
      
        error_text$ = "Blocking call canceled with WSACancelBlocking"
        
      Case #WSAEINPROGRESS
      
        error_text$ = "Socket already in use, unable to send"
        
      Case #WSAENOTSOCK
      
        error_text$ = "Not a socket!"
        
      Case #WSAESHUTDOWN
      
        error_text$ = "Socket closed."
        
      Case #WSAEOPNOTSUPP
      
        error_text$ = "OOB deteced but socket is not SOCK_STREAM type"
        
      Case #WSAEMSGSIZE
      
        error_text$ = "Datagram too big, it has been truncated"
        
      Case #WSAEINVAL
      
        error_text$ = "Socket not bound."
        
      Case #WSAECONNABORTED
      
        error_text$ = "Connection aborted due to timeout or other failure."
        
      Case #WSAECONNRESET
      
        error_text$ = "Server closed the connection."
        
      Case #WSAEWOULDBLOCK
      
        bytes_sent.l = 0
        
        total_bytes_sent.l = 0
        
        Delay(250)
        
    EndSelect

  EndIf ; = SOCKET_ERROR
  
  ;
  ; Loop until we're finished sending or we come up with an error.
  ;
  
  Until total_bytes_sent.l >= text_length.l Or bytes_sent = #SOCKET_ERROR

  ;
  ; Free up the memory bank used by our local buffer above
  ;

  FreeMemory(#LOCALBANK)
  
  ;
  ; Let the user know their send failed
  ;

  MessageRequester("Data Send Error",error_text$,0)
  
  Debug "Failed to send the following data because of this error: "+ error_text$
  
  Debug send_text$
  
  ProcedureReturn total_bytes_sent.l
  
EndProcedure

;
; Open up our window..
;

window_handle.l = Window_wnd_windowmain()

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; VERY IMPORTANT -  If you don't do this then nothing will work in your callback!
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

SetWindowCallback(@my_window_callback())

;
; Init the network and start the server listening on #PORT_NUM
;

InitNetwork()

;
; Initial state for the stop button is disabled. We can't stop before we start!
;

DisableGadget(#Gadget_wnd_windowmain_btn_stop, 1)

;
; Loop here until the window is closed.
;

Repeat
  
  EventID=WaitWindowEvent()
    
  Select EventID
    
    Case #PB_EventGadget
        
      Select EventGadgetID()
        
        Case #Gadget_wnd_windowmain_btn_stop
        
          stop_server()

          ;
          ; Now we can start the server
          ;
          
          DisableGadget(#Gadget_wnd_windowmain_btn_start, 0)
          
          ;
          ; But we can't stop it again!
          ;

          DisableGadget(#Gadget_wnd_windowmain_btn_stop, 1)


        Case #Gadget_wnd_windowmain_btn_start
        
          start_server()
          
          ;
          ; Now we can stop the server
          ;
          
          DisableGadget(#Gadget_wnd_windowmain_btn_stop, 0)
          
          ;
          ; But we can't start it again!
          ;

          DisableGadget(#Gadget_wnd_windowmain_btn_start, 1)
      
      EndSelect
         
    EndSelect

Until EventID = #PB_EventCloseWindow

;
; Stop server on exit! 
;

stop_server()

End
; ExecutableFormat=Windows
; CursorPosition=3
; FirstLine=1
; EnableOnError
; EOF